unit RTLPatching;

{$ifdef Ver100} { Delphi 3.0x }
  {$define Delphi3}
{$endif}
{$ifdef Ver110} { C++ Builder 3.0x }
  {$define Delphi3}
{$endif}
{$ifdef Ver120} { Delphi 4.0x }
  {$define Delphi4}
{$endif}
{$ifdef Ver125} { C++Builder 4.0x }
  {$define Delphi4}
{$endif}
{$ifdef ConditionalExpressions}
  {$define Delphi6AndAbove}
{$endif}

{$ifdef Delphi3}
  'This unit requires Delphi 4 or later'
{$endif}

//Define one of these two symbols to decide what to reimplement
{$define HookAssertions}
{.$define HookAutomation}

interface

var
{$ifdef HookAutomation}
  Msg: Variant;
{$endif}
  ShowPIDAndTIDInfo: Boolean = False;

implementation

uses
{$ifdef Delphi6AndAbove}
  Variants,
{$endif}
  SysUtils, Windows, DebugSupport;

{$ifndef Delphi6AndAbove}
type
  TVarType = Word;
{$endif}

//Takes pointer to one parameter from a list of disp params
//Sets Variant to match parameter and returns address of next parameter
function DispParamToVariant(ParamType: TVarType;
  RawDispParam: Pointer; var DispParam: Variant): Pointer;
var
  ParamPtrAdjust: Integer;
  VarFlag: Boolean;
const
  ParamPtrDefaultAdjust = SizeOf(Pointer);
  atTypeMask = $7F;
  atByRef    = $80;
  varWord64  = $0015;
{$ifndef Delphi6AndAbove}
  varInt64   = $0014;
{$endif}
{$ifdef Delphi4}
  varStrArg  = $0048;
{$endif}
begin
  DispParam := Null;
  Result := RawDispParam;
  VarFlag := ParamType and atByRef > 0;
  //If parameter was passed by reference, must do extra de-reference
  if VarFlag then
    RawDispParam := Pointer(RawDispParam^);
  //Normally get to next parameter by adding 4 bytes,
  //as they either take up 4 bytes, or are represented by a pointer.
  //Parameters that are not passed by ref, but which
  //take up more than 4 bytes in the parameter list
  //have to have the rest of the occupied bytes skipped
  ParamPtrAdjust := ParamPtrDefaultAdjust;
  case ParamType and atTypeMask of
  {$ifdef Delphi6AndAbove}
    varShortint, //N.B. shortints are actually passed through as varInteger
  {$endif}
    varSmallint,
    varInteger: DispParam := Integer(RawDispParam^);
    varSingle: DispParam := Single(RawDispParam^);
    varDouble:
    begin
      DispParam := Double(RawDispParam^);
      if not VarFlag then
        //Swallow extra 4 bytes taken by this 8-byte Double
        Inc(ParamPtrAdjust, (SizeOf(Double) - ParamPtrDefaultAdjust) * Ord(not VarFlag));
    end;
    varCurrency:
    begin
      DispParam := Currency(RawDispParam^);
      if not VarFlag then
        //Swallow extra 4 bytes taken by this 8-byte Currency
        Inc(ParamPtrAdjust, SizeOf(Currency) - ParamPtrDefaultAdjust);
    end;
    varDate:
    begin
      DispParam := TDateTime(RawDispParam^);
      if not VarFlag then
        //Swallow extra 4 bytes taken by this 8-byte TDateTime
        Inc(ParamPtrAdjust, SizeOf(TDateTime) - ParamPtrDefaultAdjust);
    end;
    varOleStr: DispParam := WideString(PWideChar(RawDispParam));
    varBoolean: DispParam := WordBool(RawDispParam^);
    varVariant:
    begin
      DispParam := Variant(RawDispParam^);
      if not VarFlag then
        //Swallow extra 12 bytes taken by this 16-byte Variant
        Inc(ParamPtrAdjust, SizeOf(Variant) - ParamPtrDefaultAdjust);
    end;
  {$ifdef Delphi6AndAbove}
    varLongWord, //N.B. long words are actually passed through as varInteger
    varWord,     //N.B. words are actually passed through as varInteger
  {$endif}
    varByte: DispParam := LongWord(RawDispParam^);
  {$ifdef Delphi6AndAbove} //Delphi 6 allows Int64's in Variants
    varInt64, //Int64's are actually passed as varWord64
    varWord64: DispParam := Int64(RawDispParam^);
  {$endif}
    varStrArg: DispParam := String(PChar(RawDispParam^));
  else
    DispParam := Format('Parameter has unsupported TVarType value: %d', [ParamType])
  end;
  Inc(Integer(Result), ParamPtrAdjust);
end;

type
  { Dispatch call descriptor }
  PCallDesc = ^TCallDesc;
  TCallDesc = packed record
    CallType: Byte;
    ArgCount: Byte;
    NamedArgCount: Byte;
    ArgTypes: array[0..255] of Byte;
  end;

procedure VarDispHandler(Result: PVariant; const Instance: Variant;
  CallDesc: PCallDesc; Params: Pointer); cdecl;
var
  Msg: String;       //debug msg to output
  ParamPtr: Pointer; //rolling pointer to next parameter
  NamePtr: PChar;    //pointer to arg name
  NamedArgStart,     //arg position of 1st named argument (if any)
  ArgCount: Integer; //#arguments
  Param: Variant;    //the parameter in a Variant
begin
  Msg := '';
  NamedArgStart := CallDesc.ArgCount - CallDesc.NamedArgCount;
  //After arg types, method name and named arg names are stored
  //Position pointer on method name
  NamePtr := @CallDesc.ArgTypes[CallDesc.ArgCount];
  ParamPtr := {$ifndef Delphi6AndAbove}@{$endif}Params;
  //Loop thru parameters, concatenating them together as a string
  for ArgCount := 0 to CallDesc.ArgCount - 1 do
  begin
    ParamPtr := DispParamToVariant(
      CallDesc.ArgTypes[ArgCount], ParamPtr, Param);
    //Inject named param name in string
    if ArgCount >= NamedArgStart then
    begin
      NamePtr := NamePtr + Succ(StrLen(NamePtr));
      Param := Format('%s := %s', [NamePtr, String(Param)]);
    end;
    Msg := Msg + String(Param);
  end;
  DebugMsg(Msg, ShowPIDAndTIDInfo);
end;

procedure RerouteAutomation;
begin
  VarDispProc := @VarDispHandler;
end;

procedure AssertErrorHandler(const Msg, Filename: string;  LineNumber: Integer; ErrorAddr: Pointer);
begin
  //ErrorAddr is address Assert will return to
  //(i.e. instruction following call to Assert).
  //Subtract 1 from it to ensure it refers to part
  //of the Assert call, making it more appropriate
  DebugMsg('%s (%s, line %d, address $%x)',
    [Msg, Filename, LineNumber, Pred(Integer(ErrorAddr))],
    ShowPIDAndTIDInfo);
end;

procedure RerouteAssertions;
begin
  AssertErrorProc := @AssertErrorHandler;
end;

initialization
{$ifdef HookAssertions}
  RerouteAssertions;
{$endif}
{$ifdef HookAutomation}
  {$ifdef Delphi6AndAbove}
  TVarData(Msg).VType := varUnknown;
  {$endif}
  RerouteAutomation;
{$endif}
end.
